/***** A simple instrumenter based on Sage++ ***************************************
* Usage: 1. prepare some Fortran file, say test1.f
*        2. type   "f2dep test1.f"  to generate file "test1.dep"
*        3. prepare a file instru.proj, which only contains the name "test1.dep"
*        4. type   "instrument"  to see what is output.
*
***********************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include "sage++user.h"
static int event = 0;
SgProject *project;
SgType *SgTypeInt();
SgSymbol *entry, *eexit, *loop_entry, *loop_exit;
char *fnames[] = {"entry_event", "exit_event", "loop_entry_event", "loop_exit_event"};

#define ENTRY_EVENT 0
#define EXIT_EVENT  1
#define LOOP_ENTRY  2
#define LOOP_EXIT   3
#define AFTER  0
#define BEFORE 1

void InitSymbols(SgProject *p)
{
  SgFile *f;

  f = &(p->file(0));
  entry = new SgFunctionSymb(PROCEDURE_NAME,fnames[0],
			     *SgTypeInt(),*(f->firstStatement()));
  eexit  = new SgFunctionSymb(PROCEDURE_NAME,fnames[1],
			     *SgTypeInt(), *(f->firstStatement()));
  loop_entry= new SgFunctionSymb(PROCEDURE_NAME,fnames[2],
			      *SgTypeInt(), *(f->firstStatement()));
  loop_exit = new SgFunctionSymb(PROCEDURE_NAME,fnames[3],
			     *SgTypeInt(), *(f->firstStatement()));
}

void InsertFCallNode(int event_type, SgStatement *s, int how)
{
  SgSymbol *symb;
       
  switch (event_type) {
    case ENTRY_EVENT: symb = entry; break;
    case EXIT_EVENT:  symb = eexit; break;
    case LOOP_ENTRY:  symb = loop_entry; break;
    case LOOP_EXIT :  symb = loop_exit; break;
  }

  SgCallStmt *c;
  SgValueExp *ev;

  c = new SgCallStmt(*symb);
  ev = new SgValueExp(event);
  c->addArg(*ev);
  delete(ev);

  event++;

  switch (how) {
    case BEFORE: s->insertStmtBefore(*c); break;
    case AFTER: s->insertStmtAfter(*c); break;
  }
}

void FixLoops(SgStatement* s)
{
  SgStatement *br;
  SgForStmt *f;

  for (br = s->lexNext(); ; br = br->lexNext()) {
    if (f = isSgForStmt(br))
      if (! f->isEnddoLoop())
	f->convertLoop();
   if (br == s->lastNodeOfStmt())
     break;
  }
}

void FInstrumentSub(SgStatement* s)
{
  SgStatement *br;

  for (br = s->lexNext(); ; br = br->lexNext()) {
    switch (br->variant()){
    case FOR_NODE:
      InsertFCallNode(LOOP_ENTRY, br, BEFORE);
      InsertFCallNode(LOOP_EXIT, br->lastNodeOfStmt(), AFTER);
      break;
    default:
         break;
    }
    if (br == s->lastNodeOfStmt())
      break;
  }
}

void ProjectUnparse(SgProject *project)
{
  for (int i = 0; i < project->numberOfFiles(); i++) {
    SgFile *f;
    
    f = &(project->file(i));
    f->unparsestdout();
  }
}

void FTransform(SgProject *project) {
  for (int i=0; i< project->numberOfFiles(); i++) {
    SgFile *f;
    f = &(project->file(i));
    int num_routines;
    num_routines = f->numberOfFunctions();

    for (int j=0; j < num_routines; j++) {
      SgStatement *sub;
      SgStatement *first;
      SgStatement *last;

      sub = f->functions(j);
      first = sub->lastDeclaration();
      InsertFCallNode(ENTRY_EVENT,first,AFTER);
      FixLoops(sub);
      FInstrumentSub(sub);
      last = sub->lastExecutable();
      InsertFCallNode(EXIT_EVENT, last, AFTER);
    }
  }
}

main(int argc, char *argv[]) {
  SgProject *project;
  SgFile *file;
  int c;
  project = new SgProject ("instr.proj");
  InitSymbols(project);
  FTransform(project);  
  ProjectUnparse(project);
  return(0);
}
